/*->c.postscript */



#include "stdafx.h"

#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <stdio.h>
#include <locale.h>
#include <math.h>

#include "os.h"
#include "wimp.h"
#include "wimpt.h"
#include "werr.h"
#include "akbd.h"
#include "bbc.h"
#include "flex.h"
#include "alloc.h"
#include "transform.h"
#include "xprocess.h"

#include "wos.h"

#include "err.h"
#include "xext.h"
#include "poll.h"
#include "key.h"
#include "temp.h"
#include "colour.h"
#include "xmath.h"
#include "drag.h"
#include "conf.h"
#include "redraw.h"
#include "etc.h"
#include "bits.h"
#include "hourglass.h"
#include "dbhi.h"
#include "fsx.h"
#include "bf.h"
#include "csv.h"
#include "osprint.h"
#include "drawmod.h"
#include "drawcurves.h"
#include "deb.h"
#include "bitrev.h"
#include "sprx.h"
#include "imlib.h"
#include "xxjpeg.h"



#include "str.h"
#include "constants.h"
#include "object.h"
#include "reslink.h"

#include "geometry.h"
#include "xredraw.h"

#include "view.h"
#include "viewr.h"
#include "file.h"
#include "page.h"


#include "xxcolour.h"
#include "xcolour.h"
#include "cconv.h"
#include "colsep.h"
#include "splink.h"

#include "curve.h"


#include "prstr.h"
#include "xprint.h"
#include "prstat.h"
#include "pract.h"
#include "preg.h"


#include "psinternal.h"
#include "postscript.h"




#define PSMAXSTACK 5


static ps_statestr ps_stack[PSMAXSTACK];

ps_statestr * ps_stackp;


os_error * ps_stackinit(void)
{
 ps_stackp=ps_stack;

 ps_stackp->colourmodel=BADMODEL;
 ps_stackp->linewidth=-1;
 ps_stackp->font=-1;
 ps_stackp->fontplane=-1;
	ps_stackp->overprint=-1;

 return(NULL);
}


os_error * ps_push(void)
{
 os_error * err;
 
 if((ps_stackp-ps_stack)>=(PSMAXSTACK-1)) err=generror(0,"PS stack overflow");
 else
 {
  err=NULL;
  ps_stackp++;
  *ps_stackp=*(ps_stackp-1);
 }


 return(err);
}


os_error * ps_pop(void)
{
 os_error * err;
 
 if((ps_stackp-ps_stack)<=0) err=generror(0,"PS stack underflow");
 else
 {
  err=NULL;
  ps_stackp--;
 }


 return(err);
}


os_error * ps_rgb(double r,double g,double b)
{
 os_error * err;
 char       cr[16];
 char       cg[16];
 char       cb[16];

 err=NULL;
 
 if(ps_stackp->colourmodel!=RGBMODEL ||
    ps_stackp->foreground.rgb.r!=r ||
    ps_stackp->foreground.rgb.g!=g ||
    ps_stackp->foreground.rgb.b!=b)
 {
  xftoa(r,cr,sizeof(cr),4);
  xftoa(g,cg,sizeof(cg),4);
  xftoa(b,cb,sizeof(cb),4);

  err=oswriteprinterf("%s %s %s setrgbcolor\r\n",cr,cg,cb);
  if(!err)
  {
   ps_stackp->colourmodel=RGBMODEL;
   ps_stackp->foreground.rgb.r=r;
   ps_stackp->foreground.rgb.g=g;
   ps_stackp->foreground.rgb.b=b;
  }
 }

 return(err);
}



os_error * ps_cmyk(double c,double m,double y,double k)
{
 os_error * err;
 char       cc[16];
 char       cm[16];
 char       cy[16];
 char       ck[16];

 err=NULL;
 
 if(ps_stackp->colourmodel!=CMYKMODEL ||
    ps_stackp->foreground.cmyk.c!=c ||
    ps_stackp->foreground.cmyk.m!=m ||
    ps_stackp->foreground.cmyk.y!=y ||
    ps_stackp->foreground.cmyk.k!=k)
 {
  xftoa(c,cc,sizeof(cc),4);
  xftoa(m,cm,sizeof(cm),4);
  xftoa(y,cy,sizeof(cy),4);
  xftoa(k,ck,sizeof(ck),4);


  err=oswriteprinterf("%s %s %s %s setcmykcolor\r\n",cc,cm,cy,ck);
  if(!err)
  {
   ps_stackp->colourmodel=CMYKMODEL;
   ps_stackp->foreground.cmyk.c=c;
   ps_stackp->foreground.cmyk.m=m;
   ps_stackp->foreground.cmyk.y=y;
   ps_stackp->foreground.cmyk.k=k;
  }
 }

 return(err);
}


// overprint==1 leave other seps unchanged
// overprint==0 erase other colours

static os_error * ps_overprint(int overprint)
{
 os_error * err;

 err=NULL;
 
 if(ps_stackp->overprint!=overprint)
 {
  if(postscriptlevel>=2)
		{
   err=oswriteprinterf("%s setoverprint\r\n",overprint?"true":"false");
		}

  if(!err)
  {
   ps_stackp->overprint=overprint;
  }
 }

 return(err);
}



#define PALRGB 0x1


static os_error * ps_setindexed(int ncolours,int * pal,int flags)
{
 os_error * err;
 int        i;
 int        j;

 ps_stackp->colourmodel=IDXMODEL;

 err=oswriteprinterf("[/Indexed /DeviceRGB %d\r\n<",ncolours-1);
 if(!err)
 {
  for(j=0;j<ncolours;j+=16)
  {
   for(i=j;i<ncolours && i<(j+16);i++)
   {
				if(flags & PALRGB)
				{
     err=oswriteprinterf("%06x ",((pal[i]) & 0xFF0000)|((pal[i]) & 0xFF00)|((pal[i])&0xFF));
				}
				else
				{
     err=oswriteprinterf("%06x ",((pal[i]<<8) & 0xFF0000)|((pal[i]>>8) & 0xFF00)|((pal[i]>>24)&0xFF));
				}
    if(err) break;
   }
   if(!err) err=oswriteprinterf("\r\n");
   if(err) break;
  }
  if(!err) err=oswriteprinter(">\r\n] setcolorspace\r\n");
 }

 return(err);
}


#define PSIDX1STEP 8

static os_error * ps_setindexed1(int ncolours,int * pal,int flags)
{
 os_error * err;
 int        i;
 int        j;


 ps_stackp->colourmodel=IDXMODEL;

 err=oswriteprinter("(");
 if(!err)
 {
  for(j=0;j<ncolours;j+=PSIDX1STEP)
  {
   for(i=j;i<ncolours && i<(j+PSIDX1STEP);i++)
   {
 			if(flags & PALRGB)
 			{
     err=oswriteprinterf("\\%o\\%o\\%o",(((pal[i]) & 0xFF0000)>>16),(((pal[i]) & 0xFF00)>>8),((pal[i])&0xFF));
 			}
 			else
 			{
     err=oswriteprinterf("\\%o\\%o\\%o",((pal[i]>>8)&0xFF),(((pal[i]>>16) & 0xFF)),(((pal[i]>>24) & 0xFF)));
 			}
    if(err) break;
   }
   if(!err && j<(ncolours-PSIDX1STEP)) err=oswriteprinterf("\\\r\n");
   if(err) break;
  }
  if(!err) err=oswriteprinter(")\r\n");
 }

 return(err);
}







static os_error * ps_linewidth(int linewidth)
{
 os_error * err;

 err=NULL;
 
 if(ps_stackp->linewidth!=linewidth)
 {
  err=oswriteprinterf("%d setlinewidth\r\n",linewidth);
  if(!err)
  {
   ps_stackp->linewidth=linewidth;
  }
 }

 return(err);
}




/*****************************************************************************/



#define GRXPIX(x) (scalenr((x),400,1))
#define GRYPIX(y) (scalenr((y),400,1))
#define GRFXPIX(x) (scalenr((x),1,1))
#define GRFYPIX(y) (scalenr((y),1,1))
#define GRDXPIX(x) (scalenr((x),400,256))
#define GRDYPIX(y) (scalenr((y),400,256))




static os_error * ps_draw_expandpath(int * p,transformstr * tr,int flags)
{
 int     over;
 POINT   pt[3];

 over=0;

 while(!over)
 {
  switch(*p++)
  {
   case     DRAWEND:
                    over=1;
                    break;

   case    DRAWMOVE:
                    pt[0].x=*p++;
                    pt[0].y=*p++;
                    if(tr) tr_transformset((coordstr*)pt,1,tr);
                    if(flags & DRAWMOD_OSUNITS)
                    {
                     oswriteprinterf("%d %d moveto\r\n",GRDXPIX(pt[0].x),GRDYPIX(pt[0].y));
                    }
                    else
                    if(flags & DRAWMOD_FONTUNITS)
                    {
                     oswriteprinterf("%d %d moveto\r\n",GRFXPIX(pt[0].x),GRFYPIX(pt[0].y));
                    }
                    else
                    {
                     oswriteprinterf("%d %d moveto\r\n",(pt[0].x),(pt[0].y));
                    }

                    break;

   case    DRAWDRAW:
                    pt[0].x=*p++;
                    pt[0].y=*p++;
                    if(tr) tr_transformset((coordstr*)pt,1,tr);

                    if(flags & DRAWMOD_OSUNITS)
                    {
                     oswriteprinterf("%d %d lineto\r\n",GRDXPIX(pt[0].x),GRDYPIX(pt[0].y));
                    }
                    else
                    if(flags & DRAWMOD_FONTUNITS)
                    {
                     oswriteprinterf("%d %d lineto\r\n",GRFXPIX(pt[0].x),GRFYPIX(pt[0].y));
                    }
                    else
                    {
                     oswriteprinterf("%d %d lineto\r\n",(pt[0].x),(pt[0].y));
                    }
                    
                    break;

   case   DRAWCURVE:
                    pt[0].x=*p++;
                    pt[0].y=*p++;
                    pt[1].x=*p++;
                    pt[1].y=*p++;
                    pt[2].x=*p++;
                    pt[2].y=*p++;
                    if(tr) tr_transformset((coordstr*)pt,3,tr);
                    if(flags & DRAWMOD_OSUNITS)
                    {
                     pt[0].x=GRDXPIX(pt[0].x);
																			 	pt[0].y=GRDYPIX(pt[0].y);
																			 	pt[1].x=GRDXPIX(pt[1].x);
																			 	pt[1].y=GRDYPIX(pt[1].y);
																			 	pt[2].x=GRDXPIX(pt[2].x);
																			 	pt[2].y=GRDYPIX(pt[2].y);
																				}
                    else
                    if(flags & DRAWMOD_FONTUNITS)
                    {
                     pt[0].x=GRFXPIX(pt[0].x);
																			 	pt[0].y=GRFYPIX(pt[0].y);
																			 	pt[1].x=GRFXPIX(pt[1].x);
																			 	pt[1].y=GRFYPIX(pt[1].y);
																			 	pt[2].x=GRFXPIX(pt[2].x);
																			 	pt[2].y=GRFYPIX(pt[2].y);
                    }
                    else
                    {
                     pt[0].x=(pt[0].x);
																			 	pt[0].y=(pt[0].y);
																			 	pt[1].x=(pt[1].x);
																			 	pt[1].y=(pt[1].y);
																			 	pt[2].x=(pt[2].x);
																			 	pt[2].y=(pt[2].y);
                    }
                    oswriteprinterf("%d %d %d %d %d %d curveto\r\n",(pt[0].x),(pt[0].y),(pt[1].x),(pt[1].y),(pt[2].x),(pt[2].y));
                    break;

   case   DRAWCLOSE:
                    oswriteprinter("closepath\r\n");
                    break;
  }
 }

 return(NULL);
}




os_error * ps_draw_strokeX(int * path,int thickness,int flags,int * caps,transformstr * tr,int style,int flatness)
{
 os_error   * err;
 coordstr     cx;

 err=NULL;


 if(tr)
 {
  tr_getscale(tr,&cx);
  thickness=scale(thickness,cx.x,0x10000);

		// should be root of x + y values
//  thickness=scale(thickness,currenthdcptr->vduxres,currenthdcptr->osxres);
 }

 if(flags & DRAWMOD_OSUNITS)
 {
//  thickness/=256;
  thickness=(25*thickness)/16;
 }

 if(flags & DRAWMOD_FONTUNITS)
 {
//  thickness/=400;
 }

          err=ps_draw_expandpath(path,tr,flags);
 if(!err) err=ps_linewidth(thickness);
 if(!err) err=oswriteprinter("stroke\r\n");

 return(err);

 USE(flatness);
 USE(style);
 USE(caps);
}




os_error * ps_draw_stroke(int * path,int thickness,int flags,int * caps,transformstr * tr,int style,int flatness)
{
 os_error   * err;
 int        * dest;
 int        * dest2;
 int        * dest3;


 err=NULL;


 err=draw_flatten(path,&dest,flatness);
 if(!err)
 {
  if(style) err=draw_dash(dest,&dest2,(int*)style);
  else      dest2=dest;

  if(!err)
  {
   if(thickness)
   {
    err=draw_thicken(dest2,&dest3,thickness,caps);
   }
   else dest3=dest2;

   if(!err)
   {
    err=ps_draw_expandpath(dest3,tr,flags);

    if(thickness)
    {
					err=oswriteprinter("fill\r\n");
    }
    else
    {
              err=ps_linewidth(0);
     if(!err) err=oswriteprinter("stroke\r\n");
    }

    if(thickness) sfree((flex_ptr)&dest3);
   }

   if(style) sfree((flex_ptr)&dest2);
  }
  sfree((flex_ptr)&dest);
 }


 return(err);
}




/*

 winding rule
 bits 0,1   0 non-zero winding rule
												1	-ve winding rule
												2 even odd
												3 +ve
	bits 2-5   boundary plotting

*/



os_error * ps_draw_fill(int * path,transformstr * tr,int windingrule,int flatness,int flags)
{
 os_error   * err;

 err=NULL;

 err=ps_draw_expandpath(path,tr,flags);
 oswriteprinter(((windingrule & 0x3)==2)?"eofill\r\n":"fill\r\n");

 return(err);

 USE(flatness);
 USE(windingrule);
}




os_error * ps_gwinrectdr(int x0,int y0,int x1,int y1,int * visible)
{

 *visible=1;

 return(NULL);

	USE(x0);
	USE(x1);
	USE(y0);
	USE(y1);
}





#ifdef NEVER


string pssupport_getx(int n)
{
 return(itos(getclipcurve(n) + prpagex() - prareax0())+" 400. div ");
}


string pssupport_gety(int n)
{
 return(itos((prpagey()/400)*400 - getclipcurve(n) - prareay0())+" 400. div ");
}


void pssupport_doclip(int user)
{
 int c, n;

 n = 0;

 while(c = getclipcurve(n++))
 {
  switch(c)
  {
   case 2:
          prprints(pssupport_getx(n++) + " " + pssupport_gety(n++) + " moveto\n");
          break;
   case 8:
          prprints(pssupport_getx(n++) + " " + pssupport_gety(n++) + " lineto\n");
          break;
   case 6:
          prprints(pssupport_getx(n++) + " " + pssupport_gety(n++) + " " + pssupport_getx(n++) + " " + pssupport_gety(n++) + " " + pssupport_getx(n++) + " " + pssupport_gety(n++) + " curveto\n");
          break;
   case 5:
          prprints("closepath\n");
          break;
  }
 }
 prprints("eoclip\r\n");
 prprints("newpath\r\n");
}


#endif






os_error * ps_gwindowdr(int x0,int y0,int x1,int y1)
{
	os_error * err;

	err=NULL;

	oswriteprinterf("%d % d moveto\r\n",GRDXPIX(x0),GRDYPIX(y1));
	oswriteprinterf("%d % d lineto\r\n",GRDXPIX(x1),GRDYPIX(y1));
	oswriteprinterf("%d % d lineto\r\n",GRDXPIX(x1),GRDYPIX(y0));
	oswriteprinterf("%d % d lineto\r\n",GRDXPIX(x0),GRDYPIX(y0));
	oswriteprinter("closepath\r\n");
	oswriteprinter("eoclip\r\n");
	oswriteprinter("newpath\r\n");

 return(err);
}


os_error * ps_setclipcurve(curvestr * c,transformstr * tr,viewredrawstr * v)
{
	os_error * err;

 transformstr drtr;

 setupclipcurve(c,tr,v->file);
 
 getdrtr(&drtr,NULL,v);

 err=ps_draw_expandpath(clippath,&drtr,DRAWMOD_OSUNITS);
	oswriteprinter("eoclip\r\n");
	oswriteprinter("newpath\r\n");

 return(err);
}


os_error * ps_clearclipcurve(void)
{
	clearclipcurve();
	return(NULL);
}




os_error * ps_setclipwindow(wimp_redrawstr * redrawstr,wimp_box * box,
                                   wimp_box * grbox,void ** handle,int * visible)
{
 os_error * err;
 wimp_box   clipbox;

 err=NULL;

 *grbox=redrawstr->g;
 *visible=0;


 if(grbox->x0>box->x0) clipbox.x0=grbox->x0;
 else                  clipbox.x0=box->x0;

 if(grbox->x1<box->x1) clipbox.x1=grbox->x1;
 else                  clipbox.x1=box->x1;

 if(grbox->y0>box->y0) clipbox.y0=grbox->y0;
 else                  clipbox.y0=box->y0;

 if(grbox->y1<box->y1) clipbox.y1=grbox->y1;
 else                  clipbox.y1=box->y1;


/*
 clipbox.x0=box->x0;
 clipbox.x1=box->x1;
 clipbox.y0=box->y0;
 clipbox.y1=box->y1;
*/


 if((clipbox.x0<clipbox.x1) && (clipbox.y0<clipbox.y1))
 {
		// get current clipbox for page
		// anything visible?

  ps_saveclippath(handle);
  ps_gwindowdr(clipbox.x0*256,clipbox.y0*256,clipbox.x1*256,clipbox.y1*256);


  *visible=1;
  redrawstr->g=clipbox;
 }

 return(err);
}



os_error * ps_resetclipwindow(wimp_redrawstr * redrawstr,wimp_box * grbox,void * handle)
{
	os_error * err;

	err=NULL;

 redrawstr->g=*grbox;

 return(ps_loadclippath(handle));
}









os_error * ps_saveclippath(void ** handle)
{
 os_error * err;

	err=NULL;

	*((int*)handle)=1;

 oswriteprinter("gsave\r\n");
 ps_push();

 return(err);
}



os_error * ps_loadclippath(void * handle)
{
 os_error * err;

	err=NULL;

	if((int)handle==1)
	{
  oswriteprinter("grestore\r\n");
  ps_pop();
	}


	return(err);
}









#define MAXTINT2 (MAXTINT>>2)





os_error * ps_setcolour(colourvalue * cv,int * transparent,viewstr * view)
{
 os_error * err;
 int        pal;
 colourstr* cs;
 filestr  * file;
 quadstr    q1;
 quadstr    q2;
 int        tint;
	int        overprint;

 err=NULL;

 file=view->file;

	*transparent=0;

 if(cv->data.tintdata.tag==TINTTRANS || cv->data.tintdata.tag==TINTPAL)
 {
  pal=palfromcv(cv);
  *transparent=(pal==-1);
  if(pal!=-1)
  {
   if(view && view->process)
   {
    ((processfn3)view->process)((int)view,&pal,&pal,1,PROCESST_PAL,
                                                     PROCESST_PAL,NULL,NULL);
   }
   ps_rgb(((double)((pal>>8)&0xFF))/255.0,((double)((pal>>16)&0xFF))/255.0,((double)((pal>>24)&0xFF))/255.0); 
  }
 }
 else
 {
  if(view && view->process)		// this has to be doing seps, so don't mind RGB coming back
  {
   ((processfn3)view->process)((int)view,&pal,cv,1,PROCESST_PAL,PROCESST_COLVAL,NULL,NULL);

   *transparent=(pal==-1);
   if(pal!=-1)
   {
    ps_rgb(((double)((pal>>8)&0xFF))/255.0,((double)((pal>>16)&0xFF))/255.0,((double)((pal>>24)&0xFF))/255.0);
   }
  }
  else
  {
   findcoloursp(cv->data.tint.key,&cs,file);

   if((cs->flags & COLOURTRANS) || ((cs->flags & COLOURNOPRINT) && view->printview))
   {
 			*transparent=1;
   }
   else
   {
    if(cs->flags & COLOUROVERPRINT)
    {
     if(cv->data.tint.tint>=cs->overprintlimit) overprint=1;	// transparent
     else                                       overprint=0;	// knock out
    }
    else
    if(cv->data.tint.overprint)
    {
     overprint=1;	// transparent
    }
    else
    {
     overprint=0;	// knock out
    }

				ps_overprint(overprint);
				
				if(cs->model.no==RGBMODEL)
    {
     tint=cv->data.tint.tint>>2;

     q1.type.rgb.r=((MAXTINT2-tint)*MAXCV+tint*cs->model.data.quad.type.rgb.r)/MAXTINT2;
     q1.type.rgb.g=((MAXTINT2-tint)*MAXCV+tint*cs->model.data.quad.type.rgb.g)/MAXTINT2;
     q1.type.rgb.b=((MAXTINT2-tint)*MAXCV+tint*cs->model.data.quad.type.rgb.b)/MAXTINT2;

     ps_rgb(((double)(q1.type.rgb.r))/((double)MAXCV),
            ((double)(q1.type.rgb.g))/((double)MAXCV),
            ((double)(q1.type.rgb.b))/((double)MAXCV));
    }
    else
    if(cs->model.no==HSVMODEL)
    {
     tint=cv->data.tint.tint>>2;

     hsv2rgb(&cs->model.data.quad,&q2);

     q1.type.rgb.r=((MAXTINT2-tint)*MAXCV+tint*q2.type.rgb.r)/MAXTINT2;
     q1.type.rgb.g=((MAXTINT2-tint)*MAXCV+tint*q2.type.rgb.g)/MAXTINT2;
     q1.type.rgb.b=((MAXTINT2-tint)*MAXCV+tint*q2.type.rgb.b)/MAXTINT2;

     ps_rgb(((double)(q1.type.rgb.r))/((double)MAXCV),
            ((double)(q1.type.rgb.g))/((double)MAXCV),
            ((double)(q1.type.rgb.b))/((double)MAXCV));
    }
    else
    if(cs->model.no==CMYKMODEL)
    {
     tint=cv->data.tint.tint>>2;

     q1.type.cmyk.c=(tint*cs->model.data.quad.type.cmyk.c)/MAXTINT2;
     q1.type.cmyk.m=(tint*cs->model.data.quad.type.cmyk.m)/MAXTINT2;
     q1.type.cmyk.y=(tint*cs->model.data.quad.type.cmyk.y)/MAXTINT2;
     q1.type.cmyk.k=(tint*cs->model.data.quad.type.cmyk.k)/MAXTINT2;

    // xcconv(viewstr * view,filestr * file,quadstr * s,int stype,quadstr * d,int dtype)

     ps_cmyk(((double)(q1.type.cmyk.c))/((double)MAXCV),
             ((double)(q1.type.cmyk.m))/((double)MAXCV),
             ((double)(q1.type.cmyk.y))/((double)MAXCV),
	 											((double)(q1.type.cmyk.k))/((double)MAXCV));
    }
 		}
 	}
 }

 return(err);
}



os_error * ps_setgcol(int pal,int action)
{
	os_error * err;

 err=NULL;

 if(pal!=-1)
 {
  ps_rgb(((double)((pal>>8)&0xFF))/255.0,
         ((double)((pal>>16)&0xFF))/255.0,
         ((double)((pal>>24)&0xFF))/255.0);
 }

 return(err);
 USE(action);
}




/****************************************************************************/


static char xtab[16]=
{
 '0',
 '1',
 '2',
 '3',
 '4',
 '5',
 '6',
 '7',
 '8',
 '9',
 'A',
 'B',
 'C',
 'D',
 'E',
 'F',
};




#define MAX85STRING 65
#define MAX85LEN (MAX85STRING+8)



static char string85[MAX85LEN];

static char * p85;
static int    b85;
static int    w85;



os_error * ps_hexinit(void)
{
 p85=string85;
 return(NULL);
}


os_error * ps_hexwrite(int c)
{

 *p85++=xtab[(c>>4)&0xF];
 *p85++=xtab[c&0xF];

 if((p85-string85)>MAX85STRING)
 {
  *p85++='\r';
  *p85++='\n';
  *p85=0;
  oswriteprinter(string85);
  p85=string85;
 }

 return(NULL);
}


os_error * ps_hexfinit(void)
{

 if(p85>string85)
 {
  *p85++='\r';
  *p85++='\n';
  *p85=0;
  oswriteprinter(string85);
 }

 return(NULL);
}





#define EN(c)   (int) ((c) + '!')



static void wordout(unsigned int word,char * p)
{
 *p++=(char)((EN((word / (long)(85 * 85 * 85 * 85)))));
 word %= (long)(85 * 85 * 85 * 85);
 *p++=(char)((EN(word / (85 * 85 * 85))));
 word %= (85 * 85 * 85);
 *p++=(char)((EN(word / (85 * 85))));
 word %= (85 * 85);
 *p++=(char)((EN(word / 85)));
 word %= 85;
 *p++=(char)((EN(word)));
}




static os_error * code85init(void)
{
 p85=string85;
 b85=32;
	w85=0;
 return(NULL);
}




static os_error * code85flush(void)
{
 os_error * err;

 err=NULL;

// cprintf("code flush");

 if((w85==0))
 {
  *p85++='z';
 }
 else
 {
  wordout(w85,p85);
  p85+=5;
 }

 b85=32;
 w85=0;

 if((p85-string85)>MAX85STRING)
 {
  *p85++='\r';
  *p85++='\n';
  *p85=0;
  err=oswriteprinter(string85);
  p85=string85;
 }

 return(err);
}

static os_error * code85byte(int c)
{
 os_error * err;

 err=NULL;

 b85-=8;
 w85=w85|((c & 0xFF)<<b85);
 if(b85==0) err=code85flush();

 return(err);
}




static os_error * code85word(int word)
{
 os_error * err;

          err=code85byte(word & 0xFF);
 if(!err) err=code85byte((word>>8) & 0xFF);
 if(!err) err=code85byte((word>>16) & 0xFF);
 if(!err) err=code85byte((word>>24) & 0xFF);

 return(err);
}






static os_error * code85finit(void)
{
 os_error * err;
	int        n;

 err=NULL;

	n=4-b85/8;

//	cprintf("n=%d b85 %d",n,b85);
	*p85=0;
//	cprintf("p85 %s",string85);

	if(n==4)
	{
  err=code85flush();
	}
	else
	if(n>0)
	{
  wordout(w85,p85);
  p85+=n+1;
	}
	*p85=0;
//	cprintf("p85 %s",string85);

 *p85++='~';
 *p85++='>';
 *p85++='\r';
 *p85++='\n';
 *p85=0;
 if(!err) err=oswriteprinter(string85);
 p85=string85;

 return(err);
}


static os_error * imagecodeinit(void)
{
 if(postscriptlevel<2) return(ps_hexinit());
 else                  return(code85init());

}


static os_error * imagecodefinit(void)
{
 if(postscriptlevel<2) return(ps_hexfinit());
 else                  return(code85finit());

}


static os_error * imagecodebyte(int c)
{
 if(postscriptlevel<2) return(ps_hexwrite(c));
 else                  return(code85byte(c));
}


static os_error * imagecodeword(int word)
{
	os_error * err;

 if(postscriptlevel<2)
 {
           err=ps_hexwrite(word & 0xFF);
  if(!err) err=ps_hexwrite((word>>8) & 0xFF);
  if(!err) err=ps_hexwrite((word>>16) & 0xFF);
  if(!err) err=ps_hexwrite((word>>24) & 0xFF);
 }
 else
 {
           err=code85byte(word & 0xFF);
  if(!err) err=code85byte((word>>8) & 0xFF);
  if(!err) err=code85byte((word>>16) & 0xFF);
  if(!err) err=code85byte((word>>24) & 0xFF);
 }

	return(err);
}




/************************************************************************************/

/*
  The count n = 0..127 denotes copying the next n+1 bytes literally.
    This means that literals can be from [1..128] bytes long.
  The count n = 129..255 denotes copying the next byte (257-n) times.
    This means that runs can be from [2..128] bytes long.
  The count n = 128 denotes End Of Data.
  
  pd->base <= pd->same <= pd->bp
  
  base[0] ... bp[-1] is the currently unencoded sequence
  base[0] ... same[-1] must be expressed literally
  same[0] ... bp[-1] are a run of the same character
*/


#ifdef NEVER

static char   rle_base[256];
static char * rle_same;
static char * rle_bp;



static void rle_encodeliteral(char * base,char * same)
{
 int literallen=same-base;
//  DebugAssert( literalLen <= 128 );
 if(literallen>0)
 {
  code85byte(literallen-1);
  while(base<same)
		{
   code85byte(*base++);
		}
 }
}


static void rle_encode_all(char * base,char * same,char * bp)
{
 if((bp-same)<2)
 {
  /* can't express a run shorter than 2; incorporate into previous literal. */
  same=bp;
 }

 rle_encodeliteral(base,same);
 if((bp-same)>0)
 {
		cprintf("runlen %d",257-(bp-same));
  code85byte(bp-same);
  code85byte(*same);
 }
}


static void rle_byte(int c)
{
  
	*rle_bp=(char)c;

 if(*rle_bp==*rle_same)
 {
  /* *bp continues a run.  Runs are limited to [2..128] bytes */
  if((++rle_bp-rle_same)>127)
  {
   /* hit upper run length limit, encode both prior literal and run itself */
   rle_encode_all(rle_base,rle_same,rle_bp);
   rle_same=rle_bp=rle_base;
  }
 }
 else
 {
  /* *bp ends a run.  Run of length 1 should
      be incorporated into the previous literal.  Run
      of length 2 or greater should be encoded as
      run.
  */
  int runlen;
      
  if(((runlen=(rle_bp-rle_same))>=2) || ((rle_same-rle_base)>126))
  {
   if(runlen==1)
   {
    /* incorporate singleton run into the prior literal */
    rle_same=rle_bp;
   }

   if(rle_base<rle_same)
			{
    /* encode the prior literal */
    rle_encodeliteral(rle_base,rle_same);
				memcpy(rle_base,rle_same,rle_bp-rle_same+1);
				rle_bp-=(rle_same-rle_base);
    rle_same=rle_base;
 		}

   if(runlen>=2)
   {
    /* a run of 2 or more is never
       worse than a literal; encode the run
    */

		cprintf("runlen %d",runlen);
    code85byte((257-runlen));
    code85byte(*rle_same);
    rle_bp=rle_base;
				memcpy(rle_base,rle_bp,1);
   }
 	}

// else DebugAssert( (same + 1) == bp );
  rle_same=rle_bp++;
	}
}



static void rle_word(int word)
{
 rle_byte(word & 0xFF);
 rle_byte((word>>8) & 0xFF);
 rle_byte((word>>16) & 0xFF);
 rle_byte((word>>24) & 0xFF);
}



static void rle_finit(void)
{
 rle_encode_all(rle_base,rle_same,rle_bp);
 code85byte(128);
	code85finit();
}


static void rle_init(void)
{
	code85init();
 rle_same=rle_bp=rle_base;

}




#endif




/*************************************************************************************/

#ifdef NEVER

static char xtab[16]=
{
 '0',
 '1',
 '2',
 '3',
 '4',
 '5',
 '6',
 '7',
 '8',
 '9',
 'A',
 'B',
 'C',
 'D',
 'E',
 'F',
};





static os_error   * psrendercmyk(newsprite_header * sp,transformstr * tr,
                                                          processstr * pr)
{
 os_error         * err;
 int                xpix;
 int                ypix;
 int                xdpi;
 int                ydpi;
 int              * sp2;
 int                y;
 int              * s;
 int              * d;
 int              * s1;
 int                xbyte;
 int                word;
 char             * string;
 char             * p;
 int                x;
 int                result;


 err=NULL;

 xpix=sp->width+1;
 ypix=sp->height+1;
 xdpi=sp->selector.type.xdpi;
 ydpi=sp->selector.type.ydpi;


 xexec("pssupport_cmykprerender",&result,"IIIIIIIIII",xpix,ypix,xdpi,ydpi,
                  tr->a,tr->b,tr->c,tr->d,tr->e,tr->f);

 err=flex_alloc((flex_ptr)&sp2,xpix*sizeof(int));
 if(!err)
 {
  err=flex_alloc((flex_ptr)&string,xpix*2*4+4);
  if(!err)
  {
   s=(int*)(((char*)sp)+sp->image);
  
   for(y=0;y<ypix;y++)
   {
    if(pr->mode==PROCESSFN)
    {
     d=sp2;
     pr->process(pr->handle,d,s,xpix,PROCESST_CMYK,PROCESST_CMYK);
    }
    else
    {
     d=s;
    }

    s1=d;
    p=string;

    for(x=0;x<xpix;x++)
    {
     word=*s1++;

     xbyte=(word)&0xFF;
     *p++=xtab[xbyte>>4];
     *p++=xtab[xbyte&0xF];

     xbyte=(word>>8)&0xFF;
     *p++=xtab[xbyte>>4];
     *p++=xtab[xbyte&0xF];

     xbyte=(word>>16)&0xFF;
     *p++=xtab[xbyte>>4];
     *p++=xtab[xbyte&0xF];

     xbyte=(word>>24)&0xFF;
     *p++=xtab[xbyte>>4];
     *p++=xtab[xbyte&0xF];
    }

				*p++='\r';
    *p++='\n';
    *p++=0;

    oswriteprinter(string);

    s+=xpix;
   }
   flex_free((flex_ptr)&string);
  }
  flex_free((flex_ptr)&sp2);
 }

 xexec("pssupport_cmykpostrender",&result,"");

 return(err);
}

#endif




static os_error * psrenderdeep(newsprite_header * sp,transformstr * tr,
                                                          processstr * pr,int depth)
{
 os_error         * err;
 int                xpix;
 int                ypix;
 int                xdpi;
 int                ydpi;
 int              * sp2;
 int                y;
 int              * s;
 int              * d;
 int                word;
 int                result;
 int              * lt;
 int                x;


 err=NULL;

 if(depth==2)
 {
  xpix=(32*(sp->width+1)-sp->lbit-(31-sp->rbit))/16;
 }
 else
 {
  xpix=sp->width+1;
 }
 

 ypix=sp->height+1;
 xdpi=sp->selector.type.xdpi;
 ydpi=sp->selector.type.ydpi;

 if(postscriptlevel>1)
 {
  xexec("pssupport_deepprerender",&result,"IIIIIIIIIIII",xpix,ypix,xdpi,ydpi,depth==2?3:depth,8,
                   tr->a,tr->b,tr->c,tr->d,tr->e,tr->f);
 }
 else
 {
  xexec("pssupport_deep1prerender",&result,"IIIIIIIIIIII",xpix,ypix,xdpi,ydpi,depth==2?3:depth,8,
                   tr->a,tr->b,tr->c,tr->d,tr->e,tr->f);

 }


 err=flex_alloc((flex_ptr)&sp2,xpix*sizeof(int));
 if(!err)
 {
  if(pr->mode==PROCESSBLACK)
  {
   if(depth==4) memset(sp2,0xFF,xpix*4);
   else
   if(depth==3) memset(sp2,0x0,xpix*4);
   else
   if(depth==2) memset(sp2,0x0,xpix*2); 
  }
  else
  if(pr->mode==PROCESSWHITE)
  {
   if(depth==4) memset(sp2,0x0,xpix*4);
   else
   if(depth==3) memset(sp2,0xFF,xpix*4);
   else
   if(depth==2) memset(sp2,0xFF,xpix*2); 
  }


  
  imagecodeinit();

  s=(int*)(((char*)sp)+sp->image);
  
  for(y=0;y<ypix;y++)
  {
   if(pr->mode==PROCESSFN)
   {
    d=sp2;
    if(depth==4) pr->process(pr->handle,d,s,xpix,PROCESST_CMYK,PROCESST_CMYK,NULL,NULL);
    else
    if(depth==3) pr->process(pr->handle,d,s,xpix,PROCESST_RGB32,PROCESST_RGB32,NULL,NULL); 
    else
    if(depth==2) pr->process(pr->handle,d,s,xpix,PROCESST_RGB16,PROCESST_RGB16,NULL,NULL); 
   }
   else
   if(pr->mode==PROCESSBLACK || pr->mode==PROCESSWHITE)
   {
    d=sp2;
   }
   else
   {
    d=s;
   }


   if(depth==2)
   {
    lt=d+(sp->width+1);

    x=xpix;
    while(1)
    {
     word=*d++;

     imagecodebyte(expand16[(word & 0x1F)]);
     imagecodebyte(expand16[(word & 0x3E0)>>5]);
     imagecodebyte(expand16[(word & 0x7C00)>>10]);

     
     x--;
     if(x<=0) break;

     imagecodebyte(expand16[(word & 0x1F0000)>>16]);
     imagecodebyte(expand16[(word & 0x3E00000)>>21]);
     imagecodebyte(expand16[(word & 0x7C000000)>>26]);

     
     x--;
     if(x<=0) break;
    }
   }
   else
   if(depth==3)
   {
    lt=d+xpix;
    while(d<lt)
    {
     word=*d++;
     imagecodebyte(word & 0xFF);
     imagecodebyte((word>>8) & 0xFF);
     imagecodebyte((word>>16) & 0xFF);
    }
   }
   else
   if(depth==4)
   {
    lt=d+xpix;
    while(d<lt)
    {
     imagecodeword(*d++);
    }
   }
   s+=(sp->width+1);
  }

  imagecodefinit();

  flex_free((flex_ptr)&sp2);
 }

 if(postscriptlevel>1)
 {
  xexec("pssupport_deeppostrender",&result,"");
 }
 else
 {
  xexec("pssupport_deep1postrender",&result,"");
 }

 return(err);
}




static os_error * psrenderdeepmask(newsprite_header * sp,transformstr * tr,
                                                         processstr * pr,int depth)
{
 os_error         * err;
 int                xpix;
 int                ypix;
 int                xdpi;
 int                ydpi;
 int              * sp2;
 int                y;
 int              * s;
 int              * d;
 int                word;
 int                result;
 int              * lt;
 int                x;
 int              * sm;
 int                shift;
 int                bshift;
 int                c;
 int                bpp;
	int                mbpp;
	int                istate;
	int                mshift;
	int                mword;
	int                w;
	int                h;
	int                mwordp;
 int                mshiftp;
 int              * smp;
 int                xp;



 err=NULL;

 if(depth==2)
 {
  xpix=(32*(sp->width+1)-sp->lbit-(31-sp->rbit))/16;
  bpp=16;
 }
 else
 {
  xpix=sp->width+1;
  bpp=32;
 }
 
 mbpp=1;


 ypix=sp->height+1;
 xdpi=sp->selector.type.xdpi;
 ydpi=sp->selector.type.ydpi;



	if(postscriptlevel<3)
	{
		if(postscriptlevel==2)
		{
   xexec("pssupport_deep2maskprerender",&result,"IIIIIIIIIIII",xpix,ypix,xdpi,ydpi,depth==2?3:depth,8,
                    tr->a,tr->b,tr->c,tr->d,tr->e,tr->f);
		}
		else
		{
   xexec("pssupport_deep1maskprerender",&result,"IIIIIIIIIIII",xpix,ypix,xdpi,ydpi,depth==2?3:depth,8,
                    tr->a,tr->b,tr->c,tr->d,tr->e,tr->f);
		}

  err=flex_alloc((flex_ptr)&sp2,xpix*sizeof(int));
  if(!err)
  {
   if(pr->mode==PROCESSBLACK)
   {
    if(depth==4) memset(sp2,0xFF,xpix*4);
    else
    if(depth==3) memset(sp2,0x0,xpix*4);
    else
    if(depth==2) memset(sp2,0x0,xpix*2); 
   }
   else
   if(pr->mode==PROCESSWHITE)
   {
    if(depth==4) memset(sp2,0x0,xpix*4);
    else
    if(depth==3) memset(sp2,0xFF,xpix*4);
    else
    if(depth==2) memset(sp2,0xFF,xpix*2); 
   }
  
   code85init();

   s=(int*)(((char*)sp)+sp->image);
   sm=(int*)(((char*)sp)+sp->mask);  

   for(y=0;y<ypix;y++)
   {
  
    if(pr->mode==PROCESSFN)
    {
     d=sp2;
     if(depth==4) pr->process(pr->handle,d,s,xpix,PROCESST_CMYK,PROCESST_CMYK,NULL,NULL);
     else
     if(depth==3) pr->process(pr->handle,d,s,xpix,PROCESST_RGB32,PROCESST_RGB32,NULL,NULL); 
     else
     if(depth==2) pr->process(pr->handle,d,s,xpix,PROCESST_RGB16,PROCESST_RGB16,NULL,NULL); 
    }
    else
    if(pr->mode==PROCESSWHITE || pr->mode==PROCESSBLACK)
    {
     d=sp2;
    }
    else
    {
     d=s;
    }


    c=word=mword=0;
    shift=mshift=32;
    istate=0;

				for(x=0;x<xpix;x++)
				{
     if(mshift>=32)
     {
      mword=*sm++;
      mshift=0;
     }

     if(shift>=32)
     {
      word=*d++;
      shift=0;
     }

     if(((mword>>mshift) & 0x1) && !istate)   // start
     {

      h=1;
      w=0;
      mwordp=mword;
      mshiftp=mshift;
      smp=sm;
      for(xp=x;xp<xpix;xp++)
      {
       if(!((mwordp>>mshiftp) & 0x1)) break;

       w++;

       mshiftp+=mbpp;

       if(mshiftp>=32)
       {
        mwordp=*smp++;
        mshiftp=0;
       }
      }

						if(postscriptlevel==2)
						{
       xexec("pssupport_deep2maskstrip",&result,"IIIIIIII",x,y,w,h,xpix,ypix,depth==2?3:depth,8);
						}
						else
						{
						 xexec("pssupport_deep1maskstrip",&result,"IIIIIIII",x,y,w,h,xpix,ypix,depth==2?3:depth,8);
						}

//      cprintf("x %d y %d w %d h %d bpp %d *d %x",x,y,w,h,bpp,*d);

      c=0;
      imagecodeinit();
      istate=1;
     }

     if((mword>>mshift) & 0x1)                // not masked, output
     {
						if(depth==2)
      {
       c=(word>>shift) & 0xFFFF;
                err=imagecodebyte(expand16[(c & 0x1F)]);
       if(!err) err=imagecodebyte(expand16[(c & 0x3E0)>>5]);
       if(!err) err=imagecodebyte(expand16[(c & 0x7C00)>>10]);
      }
						else
      if(depth==3)
      {
                err=imagecodebyte(word & 0xFF);
       if(!err) err=imagecodebyte((word>>8) & 0xFF);
       if(!err) err=imagecodebyte((word>>16) & 0xFF);
      }
      else
//    if(depth==4)
      {
       err=imagecodeword(word);
      }
     }

     if(((!((mword>>mshift) & 0x1)) || x>=(xpix-1)) && istate)   // finish
     {
      err=imagecodefinit();
      istate=0;
     }

     mshift+=mbpp;
     shift+=bpp;

     if(err) break;
				}

    s+=(sp->width+1);
			}

   flex_free((flex_ptr)&sp2);
  }

  if(postscriptlevel==2)
  {
   xexec("pssupport_deep2maskpostrender",&result,"");
		}
		else
		{
   xexec("pssupport_deep1maskpostrender",&result,"");
		}
	}
 else
	{
  xexec("pssupport_deep3maskprerender",&result,"IIIIIIIIIIII",xpix,ypix,xdpi,ydpi,depth==2?3:depth,8,
                    tr->a,tr->b,tr->c,tr->d,tr->e,tr->f);

  err=flex_alloc((flex_ptr)&sp2,xpix*sizeof(int));
  if(!err)
  {
   if(pr->mode==PROCESSBLACK)
   {
    if(depth==4) memset(sp2,0xFF,xpix*4);
    else
    if(depth==3) memset(sp2,0x0,xpix*4);
    else
    if(depth==2) memset(sp2,0x0,xpix*2); 
   }
   else
   if(pr->mode==PROCESSWHITE)
   {
    if(depth==4) memset(sp2,0x0,xpix*4);
    else
    if(depth==3) memset(sp2,0xFF,xpix*4);
    else
    if(depth==2) memset(sp2,0xFF,xpix*2); 
   }

  
   code85init();

   s=(int*)(((char*)sp)+sp->image);
   sm=(int*)(((char*)sp)+sp->mask);


  
   for(y=0;y<ypix;y++)
   {
    c=word=0;
    bshift=8;
    shift=32;

   
    for(x=0;x<xpix;x++)
    {
     if(shift>=32)
     {
      word=*sm++;
      shift=0;
     }

     bshift--;

     c|=(((word>>shift) & 0x1) ^ 0x1)<<bshift;

     shift++;

     if(bshift<=0)
     {
      code85byte(c);
      bshift=8;
      c=0;
     }
    }

    if(bshift<8)
    {
     code85byte(c);
    }


   
    if(pr->mode==PROCESSFN)
    {
     d=sp2;
     if(depth==4) pr->process(pr->handle,d,s,xpix,PROCESST_CMYK,PROCESST_CMYK,NULL,NULL);
     else
     if(depth==3) pr->process(pr->handle,d,s,xpix,PROCESST_RGB32,PROCESST_RGB32,NULL,NULL); 
     else
     if(depth==2) pr->process(pr->handle,d,s,xpix,PROCESST_RGB16,PROCESST_RGB16,NULL,NULL); 
    }
    else
    if(pr->mode==PROCESSWHITE || pr->mode==PROCESSBLACK)
    {
     d=sp2;
    }
    else
    {
     d=s;
    }


    if(depth==2)
    {
     lt=d+(sp->width+1);
 
     x=xpix;
     while(1)
     {
      word=*d++;

      code85byte(expand16[(word & 0x1F)]);
      code85byte(expand16[(word & 0x3E0)>>5]);
      code85byte(expand16[(word & 0x7C00)>>10]);

      x--;
      if(x<=0) break;

      code85byte(expand16[(word & 0x1F0000)>>16]);
      code85byte(expand16[(word & 0x3E00000)>>21]);
      code85byte(expand16[(word & 0x7C000000)>>26]);

      x--;
      if(x<=0) break;
     }
    }
    else
    if(depth==3)
    {
     lt=d+xpix;
     while(d<lt)
     {
      word=*d++;
      code85byte(word & 0xFF);
      code85byte((word>>8) & 0xFF);
      code85byte((word>>16) & 0xFF);
     }
    }
    else
    if(depth==4)
    {
     lt=d+xpix;
     while(d<lt)
     {
      code85word(*d++);
     }
    }

    s+=(sp->width+1);
   }

   code85finit();

   flex_free((flex_ptr)&sp2);
  }

  xexec("pssupport_deep3maskpostrender",&result,"");
 }

 return(err);
}







static os_error * psrenderdeepsep(newsprite_header * sp,transformstr * tr,
                                                        processstr * pr,int depth)
{
 os_error         * err;
 int                xpix;
 int                ypix;
 int                xdpi;
 int                ydpi;
 int              * sp2;
 int                y;
 int              * s;
 int              * d;
 int                word;
 int                result;
 int              * lt;
 int                x;



 err=NULL;

 if(depth==2)
 {
  xpix=(32*(sp->width+1)-sp->lbit-(31-sp->rbit))/16;
 }
 else
 {
  xpix=sp->width+1;
 }
 

 ypix=sp->height+1;
 xdpi=sp->selector.type.xdpi;
 ydpi=sp->selector.type.ydpi;

 if(postscriptlevel>1)
 {
  xexec("pssupport_deepprerender",&result,"IIIIIIIIIIII",xpix,ypix,xdpi,ydpi,1,8,
                   tr->a,tr->b,tr->c,tr->d,tr->e,tr->f);
 }
 else
 {
  xexec("pssupport_deep1prerender",&result,"IIIIIIIIIIII",xpix,ypix,xdpi,ydpi,1,8,
                   tr->a,tr->b,tr->c,tr->d,tr->e,tr->f);
 }

 err=flex_alloc((flex_ptr)&sp2,xpix*sizeof(int));
 if(!err)
 {
  if(pr->mode==PROCESSBLACK)
  {
   if(depth==4) memset(sp2,0x0,xpix*4);  // opposite way around to usual
   else
   if(depth==3) memset(sp2,0x0,xpix*4);
   else
   if(depth==2) memset(sp2,0x0,xpix*2); 
  }
  else
  if(pr->mode==PROCESSWHITE)
  {
   if(depth==4) memset(sp2,0xFF,xpix*4);  // opposite way around to usual
   else
   if(depth==3) memset(sp2,0xFF,xpix*4);
   else
   if(depth==2) memset(sp2,0xFF,xpix*2); 
  }

  
  imagecodeinit();

  s=(int*)(((char*)sp)+sp->image);

  
  for(y=0;y<ypix;y++)
  {
   if(pr->mode==PROCESSFN)
   {
    d=sp2;
    if(depth==4) pr->process(pr->handle,d,s,xpix,PROCESST_RGB32,PROCESST_CMYK,NULL,NULL);
    else
    if(depth==3) pr->process(pr->handle,d,s,xpix,PROCESST_RGB32,PROCESST_RGB32,NULL,NULL); 
    else
    if(depth==2) pr->process(pr->handle,d,s,xpix,PROCESST_RGB16,PROCESST_RGB16,NULL,NULL); 
   }
   else
   if(pr->mode==PROCESSBLACK || pr->mode==PROCESSWHITE)
   {
    d=sp2;
   }
   else
   {
    d=s;
   }


   if(depth==2)
   {
    lt=d+(sp->width+1);

    x=xpix;
    while(1)
    {
     word=*d++;

     imagecodebyte(expand16[(word & 0x1F)]);

     x--;
     if(x<=0) break;

     imagecodebyte(expand16[(word & 0x1F0000)>>16]);

     x--;
     if(x<=0) break;
    }
   }
   else
   if(depth==3 || depth==4)
   {
    lt=d+xpix;
    while(d<lt)
    {
     word=*d++;
     imagecodebyte(word & 0xFF);
    }
   }

   s+=(sp->width+1);
  }

  imagecodefinit();

  flex_free((flex_ptr)&sp2);
 }

 if(postscriptlevel>1)
 {
  xexec("pssupport_deeppostrender",&result,"");
 }
 else
 {
  xexec("pssupport_deep1postrender",&result,"");
 }

 return(err);
}




static os_error * psrenderdeepsepmask(newsprite_header * sp,transformstr * tr,
                                                        processstr * pr,int depth)
{
 os_error         * err;
 int                xpix;
 int                ypix;
 int                xdpi;
 int                ydpi;
 int              * sp2;
 int                y;
 int              * s;
 int              * sm;
 int              * d;
 int                word;
 int                result;
 int              * lt;
 int                x;
 int                c;
 int                shift;
 int                bshift;
	int                bpp;
	int                mbpp;
	int                istate;
	int                mshift;
	int                mword;
	int                w;
	int                h;
	int                mwordp;
 int                mshiftp;
 int              * smp;
 int                xp;


 err=NULL;

 if(depth==2)
 {
  xpix=(32*(sp->width+1)-sp->lbit-(31-sp->rbit))/16;
		bpp=16;
 }
 else
 {
  xpix=sp->width+1;
		bpp=32;
 }

	mbpp=1;
 

 ypix=sp->height+1;
 xdpi=sp->selector.type.xdpi;
 ydpi=sp->selector.type.ydpi;

	if(postscriptlevel<3)
	{
		if(postscriptlevel==2)
		{
   xexec("pssupport_deep2maskprerender",&result,"IIIIIIIIIIII",xpix,ypix,xdpi,ydpi,1,8,
                    tr->a,tr->b,tr->c,tr->d,tr->e,tr->f);
		}
		else
		{
   xexec("pssupport_deep1maskprerender",&result,"IIIIIIIIIIII",xpix,ypix,xdpi,ydpi,1,8,
                    tr->a,tr->b,tr->c,tr->d,tr->e,tr->f);
		}

  err=flex_alloc((flex_ptr)&sp2,xpix*sizeof(int));
  if(!err)
  {
   if(pr->mode==PROCESSBLACK)
   {
    if(depth==4) memset(sp2,0x0,xpix*4); // opposite way around to usual
    else
    if(depth==3) memset(sp2,0x0,xpix*4);
    else
    if(depth==2) memset(sp2,0x0,xpix*2); 
   }
   else
   if(pr->mode==PROCESSWHITE)
   {
    if(depth==4) memset(sp2,0xFF,xpix*4); // opposite way around to usual
    else
    if(depth==3) memset(sp2,0xFF,xpix*4);
    else
    if(depth==2) memset(sp2,0xFF,xpix*2); 
   }
  
   code85init();

   s=(int*)(((char*)sp)+sp->image);
   sm=(int*)(((char*)sp)+sp->mask);  

   for(y=0;y<ypix;y++)
   {
  
    if(pr->mode==PROCESSFN)
    {
     d=sp2;
     if(depth==4) pr->process(pr->handle,d,s,xpix,PROCESST_RGB32,PROCESST_CMYK,NULL,NULL);
     else
     if(depth==3) pr->process(pr->handle,d,s,xpix,PROCESST_RGB32,PROCESST_RGB32,NULL,NULL); 
     else
     if(depth==2) pr->process(pr->handle,d,s,xpix,PROCESST_RGB16,PROCESST_RGB16,NULL,NULL); 
    }
    else
    if(pr->mode==PROCESSBLACK || pr->mode==PROCESSWHITE)
    {
     d=sp2;
    }
    else
    {
     d=s;   // better never get here, doing seps with no processing
    }


    c=word=mword=0;
    shift=mshift=32;
    istate=0;

				for(x=0;x<xpix;x++)
				{
     if(mshift>=32)
     {
      mword=*sm++;
      mshift=0;
     }

     if(shift>=32)
     {
      word=*d++;
      shift=0;
     }

     if(((mword>>mshift) & 0x1) && !istate)   // start
     {

      h=1;
      w=0;
      mwordp=mword;
      mshiftp=mshift;
      smp=sm;
      for(xp=x;xp<xpix;xp++)
      {
       if(!((mwordp>>mshiftp) & 0x1)) break;

       w++;

       mshiftp+=mbpp;

       if(mshiftp>=32)
       {
        mwordp=*smp++;
        mshiftp=0;
       }
      }

						if(postscriptlevel==2)
						{
       xexec("pssupport_deep2maskstrip",&result,"IIIIIIII",x,y,w,h,xpix,ypix,1,8);
						}
						else
						{
						 xexec("pssupport_deep1maskstrip",&result,"IIIIIIII",x,y,w,h,xpix,ypix,1,8);
						}

//      cprintf("x %d y %d w %d h %d bpp %d *d %x",x,y,w,h,bpp,*d);

      c=0;
      imagecodeinit();
      istate=1;
     }

     if((mword>>mshift) & 0x1)                // not masked, output
     {
						if(depth==2)
      {
       c=(word>>shift) & 0xFFFF;
       err=imagecodebyte(expand16[(c & 0x1F)]);
      }
						else
      {
       err=imagecodebyte(word & 0xFF);
      }
     }

     if(((!((mword>>mshift) & 0x1)) || x>=(xpix-1)) && istate)   // finish
     {
      err=imagecodefinit();
      istate=0;
     }

     mshift+=mbpp;
     shift+=bpp;

     if(err) break;
				}

    s+=(sp->width+1);
			}

   flex_free((flex_ptr)&sp2);
  }

  if(postscriptlevel==2)
  {
   xexec("pssupport_deep2maskpostrender",&result,"");
		}
		else
		{
   xexec("pssupport_deep1maskpostrender",&result,"");
		}
	}
 else
	{
  xexec("pssupport_deep3maskprerender",&result,"IIIIIIIIIIII",xpix,ypix,xdpi,ydpi,1,8,
                   tr->a,tr->b,tr->c,tr->d,tr->e,tr->f);

  err=flex_alloc((flex_ptr)&sp2,xpix*sizeof(int));
  if(!err)
  {
   if(pr->mode==PROCESSBLACK)
   {
    if(depth==4) memset(sp2,0x0,xpix*4); // opposite way around to usual
    else
    if(depth==3) memset(sp2,0x0,xpix*4);
    else
    if(depth==2) memset(sp2,0x0,xpix*2); 
   }
   else
   if(pr->mode==PROCESSWHITE)
   {
    if(depth==4) memset(sp2,0xFF,xpix*4); // opposite way around to usual
    else
    if(depth==3) memset(sp2,0xFF,xpix*4);
    else
    if(depth==2) memset(sp2,0xFF,xpix*2); 
   }
  
   code85init();

   s=(int*)(((char*)sp)+sp->image);
   sm=(int*)(((char*)sp)+sp->mask);  

   for(y=0;y<ypix;y++)
   {
    c=word=0;
    bshift=8;
    shift=32;

    for(x=0;x<xpix;x++)
    {
     if(shift>=32)
     {
      word=*sm++;
      shift=0;
     }

     bshift--;

     c|=(((word>>shift) & 0x1) ^ 0x1)<<bshift;

     shift++;

     if(bshift<=0)
     {
      code85byte(c);
      bshift=8;
      c=0;
     }
    }

    if(bshift<8)
    {
     code85byte(c);
    }



   
    if(pr->mode==PROCESSFN)
    {
     d=sp2;
     if(depth==4) pr->process(pr->handle,d,s,xpix,PROCESST_RGB32,PROCESST_CMYK,NULL,NULL);
     else
     if(depth==3) pr->process(pr->handle,d,s,xpix,PROCESST_RGB32,PROCESST_RGB32,NULL,NULL); 
     else
     if(depth==2) pr->process(pr->handle,d,s,xpix,PROCESST_RGB16,PROCESST_RGB16,NULL,NULL); 
    }
    else
    if(pr->mode==PROCESSBLACK || pr->mode==PROCESSWHITE)
    {
     d=sp2;
    }
    else
    {
     d=s;   // better never get here, doing seps with no processing
    }



    if(depth==2)
    {
     lt=d+(sp->width+1);

     x=xpix;
     while(1)
     {
      word=*d++;

      code85byte(expand16[(word & 0x1F)]);

      x--;
      if(x<=0) break;

      code85byte(expand16[(word & 0x1F0000)>>16]);

      x--;
      if(x<=0) break;
     }
    }
    else
    if(depth==3 || depth==4)
    {
     lt=d+xpix;
     while(d<lt)
     {
      word=*d++;
      code85byte(word & 0xFF);
     }
    }

    s+=(sp->width+1);
   }

   code85finit();

   flex_free((flex_ptr)&sp2);
  }

  xexec("pssupport_deep3maskpostrender",&result,"");
	}

 return(err);
}





static os_error * psrenderpalette(newsprite_header * sp,transformstr * tr,processstr * pr)
{
 os_error         * err;
 int                xpix;
 int                ypix;
 int                xdpi;
 int                ydpi;
 int                y;
 int              * s;
 int                word;
 int                result;
 int                x;
 int                bpp; // real 1,2,4,8
 int                pal[256];
 int                ncolours;
 int                i;
 int                c;
 int                shift;
 int                bshift;
 int                mask;
 int                grey;


 err=NULL;

 
 getspriteinfo(sp,&xdpi,&ydpi,&bpp); 
 getspritesize(sp,&xpix,&ypix,NULL);

 spgetpal((newsprite_header*)sp,&ncolours,pal);
 if(ncolours==0) genpalette(ln2(bpp),&ncolours,pal);
 genpal256(pal,ln2(bpp),&ncolours);


 if(pr->mode==PROCESSFN)
 {
  pr->process(pr->handle,pal,pal,ncolours,PROCESST_PAL,PROCESST_PAL,NULL,NULL);
 }
 else
 if(pr->mode==PROCESSBLACK)
 {
  for(i=0;i<ncolours;i++) pal[i]=PALBLACK;
 }
 else
 if(pr->mode==PROCESSWHITE)
	{
  for(i=0;i<ncolours;i++) pal[i]=PALWHITE;
	}

 grey=(isgreyscalepal(pal,ncolours)==1);

 if(postscriptlevel>1)
 {
  xexec("pssupport_paletteprerender0",&result,"");

  err=ps_setindexed(ncolours,pal,0);
 
  xexec("pssupport_paletteprerender",&result,"IIIIIIIIIII",xpix,ypix,xdpi,ydpi,bpp,
                   tr->a,tr->b,tr->c,tr->d,tr->e,tr->f);
 }
 else
 {
  if(grey)
  {
   xexec("pssupport_deep1prerender",&result,"IIIIIIIIIIII",xpix,ypix,xdpi,ydpi,1,bpp,
                   tr->a,tr->b,tr->c,tr->d,tr->e,tr->f);
  }
  else
  {
   xexec("pssupport_palette1prerender0",&result,"");

   err=ps_setindexed1(ncolours,pal,0);
 
   xexec("pssupport_palette1prerender",&result,"IIIIIIIIIII",xpix,ypix,xdpi,ydpi,bpp,
                    tr->a,tr->b,tr->c,tr->d,tr->e,tr->f);
  }
 }

 imagecodeinit();

 mask=(1<<bpp)-1;

//	cprintf("ps pal lbit %d xpix %d wwidth %d",sp->lbit,xpix,sp->width);


 s=(int*)(((char*)sp)+sp->image);
  
 for(y=0;y<ypix;y++)
 {
  c=0;
  word=*s++;
  shift=sp->lbit;
  bshift=8;

//		cprintf("words %d",s-(int*)(((char*)sp)+sp->image));
  
  for(x=0;x<xpix;x++)
  {
   if(shift>=32)
   {
    word=*s++;
    shift=0;
   }

   bshift-=bpp;

   c|=((word>>shift) & mask)<<bshift;

   shift+=bpp;

   if(bshift<=0)
   {
    imagecodebyte(c);
    bshift=8;
    c=0;
   }
  }

  if(bshift<8)
  {
   imagecodebyte(c);
  }
 }

 imagecodefinit();

 if(postscriptlevel>1)
 {
  xexec("pssupport_palettepostrender",&result,"");
 }
 else
 {
  if(grey)
  {
   xexec("pssupport_deep1postrender",&result,"");
  }
  else
  {
   xexec("pssupport_palette1postrender",&result,"");
  }
 }

 return(err);
}




static os_error * psrenderpalettemask(newsprite_header * sp,transformstr * tr,processstr * pr)
{
 os_error         * err;
 int                xpix;
 int                ypix;
 int                xdpi;
 int                ydpi;
 int                y;
 int              * s;
 int                word;
 int                result;
 int                x;
 int                bpp; // real 1,2,4,8
 int                pal[256];
 int                ncolours;
 int                i;
 int                c;
 int                shift;
 int                bshift;
 int                mask;
 int              * sm;
 int                mbpp;
 int                mshift;
 int                mword;
 int                istate;
 int                w;
 int                h;
 int              * smp;
 int                mwordp;
 int                mshiftp;
 int                xp;


 err=NULL;

 
 getspriteinfo(sp,&xdpi,&ydpi,&bpp); 
 getspritesize(sp,&xpix,&ypix,NULL);

 spgetpal((newsprite_header*)sp,&ncolours,pal);
 if(ncolours==0) genpalette(ln2(bpp),&ncolours,pal);
 genpal256(pal,ln2(bpp),&ncolours);


 if(pr->mode==PROCESSFN)
 {
  pr->process(pr->handle,pal,pal,ncolours,PROCESST_PAL,PROCESST_PAL,NULL,NULL);
 }
 else
 if(pr->mode==PROCESSBLACK)
 {
  for(i=0;i<ncolours;i++) pal[i]=PALBLACK;
 }
 else
 if(pr->mode==PROCESSWHITE)
	{
  for(i=0;i<ncolours;i++) pal[i]=PALWHITE;
	}



 if(postscriptlevel<3)
 {
		if(postscriptlevel==2)
		{
   xexec("pssupport_palette2maskprerender0",&result,"");
   err=ps_setindexed(ncolours,pal,0);
		}
		else
		{
   xexec("pssupport_palette1maskprerender0",&result,"");
   err=ps_setindexed1(ncolours,pal,0);
		}

  if(!err)
  {
		 if(postscriptlevel==2)
			{
    xexec("pssupport_palette2maskprerender",&result,"IIIIIIIIIII",xpix,ypix,xdpi,ydpi,bpp,
                    tr->a,tr->b,tr->c,tr->d,tr->e,tr->f);
			}
			else
			{
    xexec("pssupport_palette1maskprerender",&result,"IIIIIIIIIII",xpix,ypix,xdpi,ydpi,bpp,
                    tr->a,tr->b,tr->c,tr->d,tr->e,tr->f);
			}

   mask=(1<<bpp)-1;

   s=(int*)(((char*)sp)+sp->image);
   sm=(int*)(((char*)sp)+sp->mask);  

			if(sp->selector.type.alpha) mbpp=8;
			else                        mbpp=(sp->selector.type.type)?1:bpp;

   for(y=0;y<ypix;y++)
   {
    mword=*sm++;
    word=*s++;
				c=0;
				bshift=8;
    shift=mshift=sp->lbit;
    istate=0;
  
    for(x=0;x<xpix;x++)
    {
     if(mshift>=32)
     {
      mword=*sm++;
      mshift=0;
     }

     if(shift>=32)
     {
      word=*s++;
      shift=0;
     }

     if(((mword>>mshift) & 0x1) && !istate)   // start
     {

      h=1;
      w=0;
      mwordp=mword;
      mshiftp=mshift;
      smp=sm;
      for(xp=x;xp<xpix;xp++)
      {
       if(!((mwordp>>mshiftp) & 0x1)) break;

       w++;

       mshiftp+=mbpp;

       if(mshiftp>=32)
       {
        mwordp=*smp++;
        mshiftp=0;
       }
      }

						if(postscriptlevel==2)
						{
       xexec("pssupport_palette2maskstrip",&result,"IIIIIII",x,y,w,h,xpix,ypix,bpp);
						}
						else
						{
						 xexec("pssupport_palette1maskstrip",&result,"IIIIIII",x,y,w,h,xpix,ypix,bpp);
						}

//      cprintf("x %d y %d w %d h %d",x,y,w,h);

      bshift=8;
      c=0;
      imagecodeinit();
      istate=1;
     }

     if((mword>>mshift) & 0x1)                // not masked, output
     {
      bshift-=bpp;
      c|=((word>>shift) & mask)<<bshift;
      if(bshift<=0)
      {
       err=imagecodebyte(c);
       bshift=8;
       c=0;
      }
     }

     if(((!((mword>>mshift) & 0x1)) || x>=(xpix-1)) && istate)   // finish
     {
      if(bshift<8)
      {
       err=imagecodebyte(c);
      }
      if(!err) err=imagecodefinit();

      istate=0;
     }

     mshift+=mbpp;
     shift+=bpp;

     if(err) break;
    }

    if(err) break;
   }


   if(postscriptlevel==2)
   {
    xexec("pssupport_palette2maskpostrender",&result,"");
			}
			else
			{
    xexec("pssupport_palette1maskpostrender",&result,"");
			}
  }
 }
 else
 {
  xexec("pssupport_palette3maskprerender0",&result,"");

  err=ps_setindexed(ncolours,pal,0);
  if(!err)
  {
   xexec("pssupport_palette3maskprerender",&result,"IIIIIIIIIII",xpix,ypix,xdpi,ydpi,bpp,
                   tr->a,tr->b,tr->c,tr->d,tr->e,tr->f);

   code85init();

   mask=(1<<bpp)-1;


   s=(int*)(((char*)sp)+sp->image);
   sm=(int*)(((char*)sp)+sp->mask);  

			if(sp->selector.type.alpha) mbpp=8;
			else                        mbpp=(sp->selector.type.type)?1:bpp;

   for(y=0;y<ypix;y++)
   {
    c=0;
    word=*sm++;
    bshift=8;
    shift=sp->lbit;
  
    for(x=0;x<xpix;x++)
    {
     if(shift>=32)
     {
      word=*sm++;
      shift=0;
     }

     bshift--;

     c|=(((word>>shift) & 0x1) ^ 0x1)<<bshift;

     shift+=mbpp;

     if(bshift<=0)
     {
      err=code85byte(c);
      bshift=8;
      c=0;
     }

     if(err) break;
    }

    if(err) break;

    if(bshift<8)
    {
     err=code85byte(c);
    }

    if(err) break;
  
    c=0;
    word=*s++;
    bshift=8;
    shift=sp->lbit;
  
    for(x=0;x<xpix;x++)
    {
     if(shift>=32)
     {
      word=*s++;
      shift=0;
     }

     bshift-=bpp;

     c|=((word>>shift) & mask)<<bshift;

     shift+=bpp;

     if(bshift<=0)
     {
      err=code85byte(c);
      bshift=8;
      c=0;
     }

     if(err) break;
    }

    if(err) break;

    if(bshift<8)
    {
     err=code85byte(c);
    }

    if(err) break;
   }

  
   if(!err) err=code85finit();

   xexec("pssupport_palette3maskpostrender",&result,"");
  }
 }
 return(err);
}






os_error * ps_sprender(newsprite_header * sp,transformstr * tr,
                            processstr  * pr,int * claimed)
{
 os_error * err;
 int        type;


 err=NULL;

//	cprintf("epsrender!");

	{
//	cprintf("selector type %d image=mask %d",((newsprite_header*)sp)->selector.type.type,sp->image==sp->mask);

		// gotta do something...

// if(pr->mode!=PROCESSBLACK && pr->mode!=PROCESSWHITE && 
//    pr->isprinter && !pr->separation && printtype==PRPS &&
//    (prjobbits & PRCOLOUR))


  type=((newsprite_header*)sp)->selector.type.type;

  if(type==7)
  {
   if(sp->image==sp->mask)
   {
    if(pr->separation)
    {
     err=psrenderdeepsep((newsprite_header*)sp,tr,pr,4);
    }
    else
    {
     err=psrenderdeep((newsprite_header*)sp,tr,pr,4);
    }
   }
   else
   {
    if(pr->separation)
    {
     err=psrenderdeepsepmask((newsprite_header*)sp,tr,pr,4);
    }
    else
    {
     err=psrenderdeepmask((newsprite_header*)sp,tr,pr,4);
    }
   }
  }
		else
		if(type==6)
		{
   if(sp->image==sp->mask)
   {
    if(pr->separation)
    {
     err=psrenderdeepsep((newsprite_header*)sp,tr,pr,3);
    }
    else
    {
     err=psrenderdeep((newsprite_header*)sp,tr,pr,3);
    }
   }
   else
   {
    if(pr->separation)
    {
     err=psrenderdeepsepmask((newsprite_header*)sp,tr,pr,3);
    }
    else
    {
     err=psrenderdeepmask((newsprite_header*)sp,tr,pr,3);
    }
   }
		}
		else
		if(type==5)
		{
   if(sp->image==sp->mask)
   {
    if(pr->separation)
    {
     err=psrenderdeepsep((newsprite_header*)sp,tr,pr,2);
    }
    else
    {
     err=psrenderdeep((newsprite_header*)sp,tr,pr,2);
    }
   }
   else
   {
    if(pr->separation)
    {
     err=psrenderdeepsepmask((newsprite_header*)sp,tr,pr,2);
    }
    else
    {
     err=psrenderdeepmask((newsprite_header*)sp,tr,pr,2);
    }
   }
		}
  else
  if(type==1 || type==2 || type==3 || type==4)
  {
   if(sp->image==sp->mask)
   {
    err=psrenderpalette((newsprite_header*)sp,tr,pr);
   }
   else
   {
    err=psrenderpalettemask((newsprite_header*)sp,tr,pr);
   }
  }
  else  // old format sprite
  {
   if(sp->image==sp->mask)
   {
    err=psrenderpalette((newsprite_header*)sp,tr,pr);
   }
   else
   {
    err=psrenderpalettemask((newsprite_header*)sp,tr,pr);
   }
  }

  *claimed=1;
 }

 return(err);
}




/******************************************************************************/



static os_error * psrenderframedeep(framestr * frame,transformstr * tr,processstr * pr)
{
 os_error         * err;
 int                xpix;
 int                ypix;
 int                xdpi;
 int                ydpi;
 int              * sp2;
 int                y;
 int              * s;
 int              * d;
 int                word;
 int                result;
 int              * lt;
 int                x;

 int                bpp;
 char             * p;
 int                seps;
 int                cmyk;

 err=NULL;



 seps=(pr->separation);
 cmyk=(frame->im[IM].ipal.palclass==CL_CMYK);


 xpix=frame->im[0].xpix;
 ypix=frame->im[0].ypix;
 xdpi=frame->im[0].xdpi;
 ydpi=frame->im[0].ydpi;
 bpp=frame->im[0].bpp;

// cprintf("bpp %d cmyk %d",bpp,cmyk);

 if(postscriptlevel>1)
 {
  xexec("pssupport_deepprerender",&result,"IIIIIIIIIIII",xpix,ypix,xdpi,ydpi,seps?1:(cmyk?4:3),8,
                   tr->a,tr->b,tr->c,tr->d,tr->e,tr->f);
 }
 else
 {
  xexec("pssupport_deep1prerender",&result,"IIIIIIIIIIII",xpix,ypix,xdpi,ydpi,seps?1:(cmyk?4:3),8,
                   tr->a,tr->b,tr->c,tr->d,tr->e,tr->f);
 }

 err=flex_alloc((flex_ptr)&sp2,xpix*sizeof(int));
 if(!err)
 {
  if(pr->mode==PROCESSBLACK)
  {
   memset(sp2,0x0,xpix*4);
  }
  else
  if(pr->mode==PROCESSWHITE)
  {
   memset(sp2,0xFF,xpix*4);
  }

  
  imagecodeinit();

  
  for(y=0;y<ypix;y++)
  {
   err=imfind1w(&frame->im[0],frame->im[0].topdown?y:(ypix-y-1),&s);

   if(pr->mode!=PROCESSBLACK && pr->mode!=PROCESSWHITE)
   {
    imnormalise(&frame->im[0],&s,sp2);
   }

   if(pr->mode==PROCESSFN)
   {
    d=sp2;
    if(bpp==24) pr->process(pr->handle,d,s,xpix,PROCESST_BGR24,PROCESST_BGR24,NULL,NULL);
    else
    if(bpp==32) pr->process(pr->handle,d,s,xpix,PROCESST_BGR32,cmyk?PROCESST_CMYK:PROCESST_BGR32,NULL,NULL); 
    else
    if(bpp==15 || bpp==16) pr->process(pr->handle,d,s,xpix,PROCESST_BGR16,PROCESST_BGR16,NULL,NULL); 
   }
   else
   if(pr->mode==PROCESSBLACK || pr->mode==PROCESSWHITE)
   {
    d=sp2;
   }
   else
   {
    d=s;
   }


   if(bpp==15 || bpp==16)
   {


    x=xpix;
    while(1)
    {
     word=*d++;

     if(!seps)
     {
      imagecodebyte(expand16[(word & 0x7C00)>>10]);
      imagecodebyte(expand16[(word & 0x3E0)>>5]);
     }

     imagecodebyte(expand16[(word & 0x1F)]);
     
     x--;
     if(x<=0) break;

     if(!seps)
     {
      imagecodebyte(expand16[(word & 0x7C000000)>>26]);
      imagecodebyte(expand16[(word & 0x3E00000)>>21]);
     }

     imagecodebyte(expand16[(word & 0x1F0000)>>16]);

     
     x--;
     if(x<=0) break;
    }
   }
   else
   if(bpp==32)
   {
//    cprintf("cmyk %d",cmyk);

    lt=d+xpix;
    while(d<lt)
    {
     word=*d++;

     if(cmyk)
     {
      imagecodebyte(((word) & 0xFF));
      imagecodebyte(((word>>8) & 0xFF));
      imagecodebyte(((word>>16) & 0xFF));
      imagecodebyte(((word>>24) & 0xFF));
     }
     else
     {
      if(!seps)
      {
       imagecodebyte((word>>16) & 0xFF);
       imagecodebyte((word>>8) & 0xFF);
      }
      imagecodebyte(word & 0xFF);
     }
    }
   }
   else
   if(bpp==24)
   {
    p=(char*)d;
    for(x=0;x<xpix;x++)
    {
     if(!seps)
     {
      imagecodebyte(*(p+2));
      imagecodebyte(*(p+1));
     }
     imagecodebyte(*p);
     p+=3;
    }
   }
  }

  imagecodefinit();

  flex_free((flex_ptr)&sp2);
 }

 if(postscriptlevel>1)
 {
  xexec("pssupport_deeppostrender",&result,"");
 }
 else
 {
  xexec("pssupport_deep1postrender",&result,"");
 }

 return(err);
}



static os_error * psrenderframepalette(framestr * frame,transformstr * tr,processstr * pr)
{
 os_error         * err;
 int                xpix;
 int                ypix;
 int                xdpi;
 int                ydpi;
 int                y;
 int              * s;
 int                word;
 int                result;
 int                x;
 int                bpp; // real 1,2,4,8
 int                pal[256];
 int                ncolours;
 int                i;
 int                c;
 int                shift;
 int                pixelperbyte;
 int                grey;


 err=NULL;

 xpix=frame->im[0].xpix;
 ypix=frame->im[0].ypix;
 xdpi=frame->im[0].xdpi;
 ydpi=frame->im[0].ydpi;
 bpp=frame->im[0].bpp;
 ncolours=frame->im[0].ipal.ncolours;

 for(i=0;i<ncolours;i++)
 {
  pal[i]=frame->im[0].ipal.word[i];
 }

 if(pr->mode==PROCESSFN)
 {
  pr->process(pr->handle,pal,pal,ncolours,PROCESST_BGR32,PROCESST_BGR32,NULL,NULL);
 }
 else
 if(pr->mode==PROCESSBLACK)
 {
  for(i=0;i<ncolours;i++) pal[i]=0;
 }
 else
 if(pr->mode==PROCESSWHITE)
	{
  for(i=0;i<ncolours;i++) pal[i]=0xFFFFFF;
	}

 grey=(isgreyscalepal(pal,ncolours)==1);

 if(postscriptlevel>1)
 {
  xexec("pssupport_paletteprerender0",&result,"");

  err=ps_setindexed(ncolours,pal,PALRGB);

  xexec("pssupport_paletteprerender",&result,"IIIIIIIIIII",xpix,ypix,xdpi,ydpi,bpp,
                   tr->a,tr->b,tr->c,tr->d,tr->e,tr->f);
 }
 else
 {
  if(grey)
  {
   xexec("pssupport_deep1prerender",&result,"IIIIIIIIIIII",xpix,ypix,xdpi,ydpi,1,bpp,
                   tr->a,tr->b,tr->c,tr->d,tr->e,tr->f);
  }
  else
  {

   xexec("pssupport_palette1prerender0",&result,"");

   err=ps_setindexed1(ncolours,pal,PALRGB);
 
   xexec("pssupport_palette1prerender",&result,"IIIIIIIIIII",xpix,ypix,xdpi,ydpi,bpp,
                    tr->a,tr->b,tr->c,tr->d,tr->e,tr->f);
  }
 }


 imagecodeinit();

 pixelperbyte=8/bpp;

// we handle pixels in bytes, the pixel order is the same in PS and BMP data
// if we handled each pixel one by one we'd have to extract them in the
// correct order, first pixel in byte from topmost bits and so on.

 for(y=0;y<ypix;y++)
 {
  err=imfind1w(&frame->im[0],frame->im[0].topdown?y:(ypix-y-1),&s);
  
  word=0;
  shift=32;
  
  for(x=0;x<xpix;x+=pixelperbyte)
  {
   if(shift>=32)
   {
    word=*s++;
    shift=0;
   }

   c=((word>>shift) & 0xFF);

   shift+=8;

   imagecodebyte(c);
  }

//  cprintf("c=%x word %x bshift %d shift %d",c,word,bshift,shift);
 }

 imagecodefinit();


 if(postscriptlevel>1)
 {
  xexec("pssupport_palettepostrender",&result,"");
 }
 else
 {
  if(grey)
  {
   xexec("pssupport_deep1postrender",&result,"");
  }
  else
  {
   xexec("pssupport_palette1postrender",&result,"");
  }
 }

 return(err);
}





os_error * ps_renderframe(framestr * frame,transformstr * tr,processstr * pr)
{
 os_error * err;

 err=NULL;

 if(frame->im[0].bpp>8)
 {
  err=psrenderframedeep(frame,tr,pr);
 }
 else
 {
  err=psrenderframepalette(frame,tr,pr);
 }

 return(err);
}





/*****************************************************************************/




os_error * ps_jprender(char * data,int len,transformstr * tr,
                            processstr  * pr,int * claimed)
{
 os_error    * err;
 char        * p;
 char        * lt;
 jpeginfostr   info;
 int           result;

 
 
 err=NULL;

//	cprintf("epsrender!");

	{
  if(pr->mode==PROCESSNONE && postscriptlevel>1)
  {
   err=jpeg_info(data,len,&info);

//			cprintf("jprender >>> xdpi %d ydpi %d",info.xdpi,info.ydpi);

   xexec("pssupport_jpegprerender",&result,"IIIIIIIIIII",info.xpix,info.ypix,info.xdpi,info.ydpi,info.components,
                  tr->a,tr->b,tr->c,tr->d,tr->e,tr->f);

  
   code85init();

   p=data;
   lt=p+len;
   while(p<lt)
   {
    code85byte(*p++);
   }

   code85finit();

   xexec("pssupport_jpegpostrender",&result,"");
  }
  else
  {
   framebasestr  * framebase;
   framestr      * frames;

   err=jpeg_getframe(data,len,&framebase);
   if(!err)
   {
    frames=framebase->frame;
    ps_renderframe(frames,tr,pr);
    imscrapframebase(&framebase);
   }
  }

  *claimed=1;
 }

 return(err);
}











